"""
This module pre-defines colors from FlatUI.
https://materialui.co/flatuicolors/
"""

from arcade.color import BLACK, WHITE, Color

WHITE = WHITE
GREEN_TURQUOISE = Color(26, 188, 156, 255)
GREEN_GREEN_SEA = Color(22, 160, 133, 255)
GREEN_NEPHRITIS = Color(39, 174, 96, 255)
GREEN_EMERALD = Color(46, 204, 113, 255)
BLUE_PETER_RIVER = Color(52, 152, 219, 255)
BLUE_BELIZE_HOLE = Color(41, 128, 185, 255)
PURPLE_AMETHYST = Color(155, 89, 182, 255)
PURPLE_WISTERIA = Color(142, 68, 173, 255)
DARK_BLUE_WET_ASPHALT = Color(52, 73, 94, 255)
DARK_BLUE_MIDNIGHT_BLUE = Color(44, 62, 80, 255)
YELLOW_SUN_FLOWER = Color(241, 196, 15, 255)
YELLOW_ORANGE = Color(243, 156, 18, 255)
ORANGE_CARROT = Color(230, 126, 34, 255)
ORANGE_PUMPKIN = Color(211, 84, 0, 255)
RED_ALIZARIN = Color(231, 76, 60, 255)
RED_POMEGRANATE = Color(192, 57, 43, 255)
WHITE_CLOUDS = Color(236, 240, 241, 255)
WHITE_SILVER = Color(189, 195, 199, 255)
GRAY_CONCRETE = Color(149, 165, 166, 255)
GRAY_ASBESTOS = Color(127, 140, 141, 255)
BLACK = BLACK

__all__ = [
    "BLACK",
    "WHITE",
    "GREEN_TURQUOISE",
    "GREEN_GREEN_SEA",
    "GREEN_NEPHRITIS",
    "GREEN_EMERALD",
    "BLUE_PETER_RIVER",
    "BLUE_BELIZE_HOLE",
    "PURPLE_AMETHYST",
    "PURPLE_WISTERIA",
    "DARK_BLUE_WET_ASPHALT",
    "DARK_BLUE_MIDNIGHT_BLUE",
    "YELLOW_SUN_FLOWER",
    "YELLOW_ORANGE",
    "ORANGE_CARROT",
    "ORANGE_PUMPKIN",
    "RED_ALIZARIN",
    "RED_POMEGRANATE",
    "WHITE_CLOUDS",
    "WHITE_SILVER",
    "GRAY_CONCRETE",
    "GRAY_ASBESTOS",
]
